"""Class used to represent and handle table/chart objects."""


class MatrixTable:
    """Class for handling tables and charts in a functional structure, allowing algorithmic correction and reading.

    Fields:
        title: Title of the table, empty by default.
        content: Matrix representation of the table content.
        is_chart: Boolean marker indicating whether the data table must be read as a chart.
    """

    def __init__(
        self: "MatrixTable",
        title: str,
        content: list[list[str]],
        is_chart: bool = False,
    ) -> None:
        """Create a MatrixTable object."""
        self.title: str = title
        self.content: list[list[str]] = content
        self.is_chart: bool = is_chart

    def abort(self: "MatrixTable") -> bool:
        """Returns a boolean value indicating whether the table is smaller than 2x2, in which case must be read as a string."""
        return len(self.content) <= 1 or len(self.content[0]) <= 1

    def read(self: "MatrixTable") -> str:
        """Returns a string representing the table/chart with its title and content."""
        if self.is_chart:
            return (
                "** Chart : "
                + str(self.title)
                + " "
                + str(self.content)
                + " ; End Of Chart ** "
            )
        return (
            "** Table : "
            + str(self.title)
            + " "
            + str(self.content)
            + " ; End Of Table ** "
        )

    def read_as_str(self: "MatrixTable") -> str:
        """Returns a string containing all the text in the table, without any structure."""
        text = self.title + " "
        for i in range(len(self.content)):
            for j in range(len(self.content[0])):
                text += self.content[i][j] + " "
        return text



"""Extract subchunks from PDF file."""

import base64
import operator
from functools import reduce
from pathlib import Path
from typing import Any, Callable, List, Tuple

import fitz
import pymupdf

from unichunking.types import ChunkPosition, SubChunk
from unichunking.utils import logger

MIN_PIXELS: int = 128


def _handle_line(
    line: Any,
    width: float,
    height: float,
    subchunk_idx: int,
    page_num: int,
    file_name: str,
) -> Tuple[List[SubChunk], int]:
    line_chunks: List[SubChunk] = []
    for span in line["spans"]:
        text = str(span["text"].replace("�", " ").strip())
        font: Any = span["font"]
        bbox: Any = span["bbox"]
        if "bold" in font.lower():
            text = f"**{text}**"
        if text:
            x0, y0, x1, y1 = bbox
            position = ChunkPosition(
                x0=x0 / width,
                y0=y0 / height,
                x1=x1 / width,
                y1=y1 / height,
            )
            line_chunks.append(
                SubChunk(
                    subchunk_id=subchunk_idx,
                    content=text,
                    page=page_num,
                    position=position,
                    file_name=file_name,
                ),
            )
            subchunk_idx += 1

    return line_chunks, subchunk_idx


async def _retrieve_subchunks(  # noqa: C901, PLR0912
    path: Path,
    status_manager: Any,
    function: Callable[[str], str],
) -> Tuple[List[List[List[List[SubChunk]]]], List[SubChunk]]:
    chunks: List[List[List[List[SubChunk]]]] = []
    images_chunks: List[SubChunk] = []
    idx = 0
    image_idx = 0

    with pymupdf.Document(path) as doc:
        num_pages: Any = doc.page_count  # type: ignore
        for page_num in range(num_pages):
            if page_num % int(num_pages / 17 + 1) == 0:
                page_progress = int((page_num + 1) / num_pages * 75)
                await status_manager.update_status(
                    progress=page_progress,
                    start=status_manager.start,
                    end=status_manager.end,
                )
            page_chunks: List[List[List[SubChunk]]] = []
            textpage: Any = doc.load_page(page_num).get_textpage()  # type: ignore
            page = textpage.extractDICT(sort=False)
            width, height = page["width"], page["height"]
            blocks = page["blocks"]

            for block in blocks:
                block_chunks: List[List[SubChunk]] = []
                lines = block["lines"]
                for line in lines:
                    line_chunks, idx = _handle_line(
                        line=line,
                        width=width,
                        height=height,
                        subchunk_idx=idx,
                        page_num=page_num,
                        file_name=path.name,
                    )
                    if line_chunks:
                        block_chunks.append(line_chunks)
                if block_chunks:
                    page_chunks.append(block_chunks)
            if page_chunks:
                chunks.append(page_chunks)

        if idx != 0:  # mean we found some text
            for page_num in range(num_pages):
                doc_page = doc.load_page(page_num)  # type: ignore
                for image in doc_page.get_images(full=True):  # type: ignore
                    rect = doc_page.get_image_bbox(image)  # type: ignore
                    x0, y0, x1, y1 = rect.x0, rect.y0, rect.x1, rect.y1  # type: ignore
                    image_width, image_height = int(image[2]), int(image[3])  # type: ignore

                    if image_width >= MIN_PIXELS and image_height >= MIN_PIXELS:
                        image_bytes = doc.extract_image(image[0])["image"]  # type: ignore
                        try:
                            base64_image = base64.b64encode(image_bytes).decode("utf-8")  # type: ignore
                            image_description = function(base64_image)
                            image_chunk = SubChunk(
                                subchunk_id=7_700_000 + image_idx,
                                content=image_description,
                                page=page_num,
                                position=ChunkPosition(
                                    x0=x0 / width,  # type: ignore
                                    y0=y0 / height,  # type: ignore
                                    x1=x1 / width,  # type: ignore
                                    y1=y1 / height,  # type: ignore
                                ),
                                file_name=path.name,
                                content_type="image",
                            )
                            image_idx += 1
                            images_chunks.append(image_chunk)
                        except Exception as e:  # noqa: BLE001
                            logger.debug(f"Image couldn't be processed : {e}")

        else:
            for page_num in range(num_pages):
                # we will just screen shot the whole pages
                page = doc.load_page(page_num)  # type: ignore

                # Définir la résolution
                zoom = 1.0  # Modifier le facteur de zoom pour une meilleure qualité
                mat = fitz.Matrix(zoom, zoom)

                # Convertir la page en image pixmap
                pix = page.get_pixmap(matrix=mat)  # type: ignore
                image_bytes = pix.tobytes()  # type: ignore

                base64_image = base64.b64encode(image_bytes).decode("utf-8")  # type: ignore
                image_description = function(base64_image)
                image_chunk = SubChunk(
                    subchunk_id=7_700_000 + image_idx,
                    content=image_description,
                    page=page_num,
                    position=ChunkPosition(
                        x0=0,
                        y0=0,
                        x1=1,
                        y1=1,
                    ),
                    file_name=path.name,
                    content_type="image",
                )
                image_idx += 1
                images_chunks.append(image_chunk)

    return chunks, images_chunks


def _filter_subchunks(
    chunks: List[List[List[List[SubChunk]]]],
) -> List[SubChunk]:
    flattened_chunks: List[SubChunk] = []

    for page_chunks in chunks:
        for block_chunks in page_chunks:
            for line_chunks in block_chunks:
                if line_chunks:
                    filtered_line_chunks = reduce(operator.add, line_chunks)
                    flattened_chunks.append(filtered_line_chunks)

    return flattened_chunks


async def extract_subchunks_pdf(
    path: Path,
    status_manager: Any,
    function: Callable[[str], str],
) -> List[SubChunk]:
    """Filetype-specific function : extracts subchunks from a PDF file.

    Args:
        path: Path to the local file.
        status_manager: Optional, special object to manage task progress.
        function: Function to handle images.

    Returns:
        A list of SubChunk objects.
    """
    chunks, images_chunks = await _retrieve_subchunks(
        path=path,
        status_manager=status_manager,
        function=function,
    )
    flattened_chunks = _filter_subchunks(chunks)

    flattened_chunks.extend(images_chunks)

    progress = 100
    await status_manager.update_status(
        progress=progress,
        start=status_manager.start,
        end=status_manager.end,
    )

    return flattened_chunks



"""Reads MarkDown and identifies tables."""

from unichunking.settings import unisettings
from unichunking.types import ChunkPosition, MatrixTable, SubChunk


def md_interpreter(
    lines: list[str],
) -> tuple[list[SubChunk], list[MatrixTable], list[str]]:
    """Reads MarkDown with the ability to identify tables.

    Args:
        lines: Lines of text extracted from file.

    Returns:
        A tuple containing three lists:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
    """
    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []

    i_line = 0
    while i_line < len(lines):
        if any(
            table_marker in lines[i_line]
            for table_marker in unisettings.text.TABLE_MARKERS
        ):
            if len(subchunks) > 0:
                subchunks.pop()
            table: list[list[str]] = []
            line_idx = i_line - 1
            while 0 <= line_idx < len(lines) and "|" in lines[line_idx]:
                if line_idx == i_line:
                    line_idx += 1
                    continue
                line_text = lines[line_idx].replace("\n", "").strip()
                line_content = line_text.split("|")
                if line_text[0] == "|":
                    line_content.pop(0)
                if line_text[-1].strip() == "|":
                    line_content.pop()
                table.append([cell.strip() for cell in line_content])
                line_idx += 1
            subchunks.append(
                SubChunk(
                    subchunk_id=0,
                    content=f"{len(tables)}",
                    page=0,
                    position=ChunkPosition(0, 0, 0, 0),
                    file_name="",
                    content_type="table",
                ),
            )
            tables.append(MatrixTable("", table))
            i_line = line_idx

        else:
            subchunk_content = lines[i_line].strip() + " "
            if subchunk_content:
                subchunks.append(
                    SubChunk(
                        subchunk_id=0,
                        content=subchunk_content + " ",
                        page=0,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name="",
                    ),
                )
            i_line += 1

    return subchunks, tables, images



"""Settings for universal chunking."""

from pydantic_settings import BaseSettings

from unichunking.settings.chunking import ChunkingSettings
from unichunking.settings.files import FilesSettings
from unichunking.settings.text import TextSettings


class UniSettings(BaseSettings):
    """Base settings for file processing."""

    chunking: ChunkingSettings = ChunkingSettings()
    files: FilesSettings = FilesSettings()
    text: TextSettings = TextSettings()


unisettings = UniSettings()



"""Functions to extract data from XML structures."""

from pathlib import Path
from zipfile import ZipFile

from unichunking.types import MatrixTable


def _clean_xml(string: str) -> str:
    return (
        string.replace("\\xc3\\xa9", "é").replace("&apos;", "'").replace("&#10;", "\n")
    )


def _extract_charts_xml(
    file_path: Path,
    element_xml: str,
    extension: str,
    page_idx: int,
) -> list[str]:
    """Extracts the XML encoding of a chart in a DOCX/PPTX file.

    Args:
        file_path: Path to the local file.
        element_xml: XML encoding of the element pointing to the chart in the file.
        extension: File extension.
        page_idx: Index of the page containing the chart in the file.

    Returns:
        A list of length 1 (for compatibility reasons) containing the XML encoding of the chart.
    """
    with file_path.open("rb") as doc:
        zip_doc = ZipFile(doc)

        if extension == "docx":
            app = "word"
            rel_path = "word/_rels/document.xml.rels"

        else:
            app = "ppt"
            rel_path = f"ppt/slides/_rels/slide{page_idx}.xml.rels"

        element_charts = element_xml.split('r:id="rId')[1:]
        charts_ids = [element_chart.split('"/>')[0] for element_chart in element_charts]
        charts_paths = [
            (
                str(zip_doc.read(rel_path))
                .split(f'<Relationship Id="rId{chart_id}"')
                .pop()
                .split('"/>')[0]
                .split('Target="')
                .pop()
            ).replace("../", "")
            for chart_id in charts_ids
        ]

        return [
            str(
                zip_doc.read(
                    f"{app}/{chart_path}",
                ),
            )
            for chart_path in charts_paths
        ]


def extract_charts(
    file_path: Path,
    element_xml: str,
    extension: str,
    page_idx: int = 0,
    charts_xml: list[str] | None = None,
) -> list[MatrixTable]:
    """Extracts the data of Office charts from their XML encoding.

    Args:
        file_path: Path to the local file.
        element_xml: XML encoding of the element pointing to the chart in the file.
        extension: File extension.
        page_idx: Index of the page containing the chart in the file.
        charts_xml: A list containing the XML encodings of all charts to process.

    Returns:
        A list of chart objects, of class MatrixTable.
    """
    if charts_xml is None:
        charts_xml = _extract_charts_xml(
            file_path=file_path,
            element_xml=element_xml,
            extension=extension,
            page_idx=page_idx,
        )

    tables: list[MatrixTable] = []

    for chart_xml in charts_xml:
        title = chart_xml.split("<a:t>").pop().split("</a:t>")[0]
        title = (title != chart_xml) * _clean_xml(title)

        plots = chart_xml.split("<c:plotArea>")
        plots.pop(0)

        for plot in plots:
            table = [[""]]
            categories = (
                plot.split("<c:strCache>")
                .pop()
                .split("</c:strCache>")[0]
                .split("<c:v>")
            )
            for category in categories[1:]:
                table[0].append(_clean_xml(category.split("</c:v>")[0]))
            series = plot.split("<c:strCache>")
            for i_series in range(1, len(series)):
                if i_series % 2:
                    table.append(
                        [
                            _clean_xml(
                                series[i_series]
                                .split("<c:v>")
                                .pop()
                                .split("</c:v>")[0],
                            ),
                        ],
                    )
                else:
                    table[-1].extend(
                        [
                            value.split("</c:v>")[0]
                            for value in series[i_series]
                            .split("</c:strCache")
                            .pop()
                            .split("<c:v>")[1:]
                        ],
                    )

            tables.append(MatrixTable(title, table, is_chart=True))

    return tables


def extract_media_xlsx(
    file_path: Path,
    page_idx: int,
) -> tuple[list[MatrixTable], list[bytes]]:
    """Extracts all charts and images on a given sheet of an XLSX file.

    Args:
        file_path: Path to the local file.
        page_idx: Index of the sheet to process.

    Returns:
        A tuple containing two lists:
        - List of table/chart objects.
        - List of image objects.
    """
    charts_xml: list[str] = []
    images: list[bytes] = []

    with file_path.open("rb") as doc:
        zip_doc = ZipFile(doc)

        drawings_rels = str(zip_doc.read(f"xl/worksheets/sheet{page_idx}.xml")).split(
            '<drawing r:id="rId',
        )[1:]
        if not drawings_rels:
            return [], []
        sheet_rels = str(zip_doc.read(f"xl/worksheets/_rels/sheet{page_idx}.xml.rels"))

        drawings: list[str] = []
        for drawing_rel in drawings_rels:
            drawing_rid = drawing_rel.split('"/>')[0]
            drawings.append(
                sheet_rels.split(f'<Relationship Id="rId{drawing_rid}')
                .pop()
                .split('"/>')[0]
                .split('Target="../drawings/')
                .pop(),
            )

        for drawing in drawings:
            drawing_xml = str(zip_doc.read(f"xl/drawings/{drawing}"))
            drawing_rels = str(zip_doc.read(f"xl/drawings/_rels/{drawing}.rels"))

            drawing_charts = drawing_xml.split(
                "<c:chart ",
            )[1:]
            charts_rid = [
                chart.split('r:id="rId')[1].split('"/>')[0] for chart in drawing_charts
            ]
            charts_paths = [
                drawing_rels.split(f'<Relationship Id="rId{chart_rid}')
                .pop()
                .split('Target="../')[1]
                .split('"/>')[0]
                for chart_rid in charts_rid
            ]
            charts_xml.extend(
                [str(zip_doc.read(f"xl/{chart_path}")) for chart_path in charts_paths],
            )

            drawing_images = drawing_xml.split(
                '<a:blip r:embed="rId',
            )[1:]
            images_rid = [image.split('">')[0] for image in drawing_images]
            images_paths = [
                drawing_rels.split(f'<Relationship Id="rId{image_rid}')
                .pop()
                .split('Target="../')[1]
                .split('"/>')[0]
                for image_rid in images_rid
            ]
            images.extend(
                [zip_doc.read(f"xl/{image_path}") for image_path in images_paths],
            )

        return (
            extract_charts(
                file_path=file_path,
                element_xml="",
                extension="xlsx",
                page_idx=page_idx,
                charts_xml=charts_xml,
            ),
            images,
        )



"""Extract subchunks from XLSX file."""

from pathlib import Path
from typing import Any

import openpyxl
from xlcalculator import Evaluator, ModelCompiler

from unichunking.tools import (
    compact_matrix,
    convert_file,
    extract_media_xlsx,
)
from unichunking.types import (
    ChunkPosition,
    MatrixTable,
    SubChunk,
)
from unichunking.utils import logger


def _is_empty_line(table: list[list[str]], i: int) -> bool:
    if (not table) or i >= len(table):
        return False
    return all(not table[i][j] for j in range(len(table[i])))


def _is_empty_column(table: list[list[str]], j: int) -> bool:
    if (not table) or j >= len(table[0]):
        return False
    return all(not table[i][j] for i in range(len(table)))


def _clean_borders(table: list[list[str]]) -> list[list[str]]:
    keep_going = len(table) > 0
    while keep_going:
        if _is_empty_line(table, 0):
            table.pop(0)
        elif _is_empty_line(table, len(table) - 1):
            table.pop()
        elif _is_empty_column(table, 0):
            for i in range(len(table)):
                table[i].pop(0)
        elif _is_empty_column(table, len(table) - 1):
            for i in range(len(table)):
                table[i].pop()
        else:
            keep_going = False
        keep_going = keep_going and len(table) > 0 and len(table[0]) > 0
    return table


def _extract_table_from_sheet(
    wb: openpyxl.Workbook,
    sheet: Any,
    sheet_idx: int,
    evaluator: Evaluator,
) -> MatrixTable:
    """Extracts the content of a XLSX sheet in the form of the minimal rectangular area that contains all non-empty cells on the sheet.

    Formulae are replaced by their outputs.

    Args:
        wb: The Workbook object representing the active file.
        sheet: The Worksheet object representing the active sheet.
        sheet_idx: The index of the active sheet.
        evaluator: The Evaluator object used to compute formulae outputs.

    Returns;
        The MatrixTable corresponding to the minimal rectangular area that contains all non-empty cells on the sheet.
    """
    merged_cells = sheet.merged_cells.ranges.copy()
    for merge_range in merged_cells:
        value = sheet.cell(merge_range.min_row, merge_range.min_col).value
        sheet.unmerge_cells(str(merge_range))
        for row in sheet.iter_rows(
            min_row=merge_range.min_row,
            max_row=merge_range.max_row,
            min_col=merge_range.min_col,
            max_col=merge_range.max_col,
        ):
            for cell in row:
                cell.value = value

    table: list[list[str]] = []
    for row_idx in range(sheet.min_row, sheet.max_row + 1):
        line: list[str] = []
        for col_idx in range(sheet.min_column, sheet.max_column + 1):
            # Computing the result of a formula :
            if sheet.cell(row_idx, col_idx).data_type == "f":
                cell_adress = f"{wb.sheetnames[sheet_idx]}!{sheet.cell(row_idx, col_idx).coordinate}"
                try:
                    cell_value: str = str(evaluator.evaluate(cell_adress))  # type: ignore
                    sheet.cell(row_idx, col_idx).value = cell_value
                except Exception as e:  # noqa: BLE001
                    logger.debug(
                        f"Error occured when computing formula : {e}",
                    )

            line.append(str(sheet.cell(row_idx, col_idx).value).replace("None", ""))

        table.append(line)

    return MatrixTable("", table)


def _split_xlsx_sheet(sheet_table: MatrixTable) -> list[MatrixTable]:
    """Identifies the different groups of cells on a XLSX sheet, initially represented by a single table containing all the non-empty cells.

    Iteratively removes blank lines/columns on the edges and splits the table along interior blank lines/columns.

    Args:
        sheet_table: A MatrixTable object representing the minimal rectangular area that contains all non-empty cells on the sheet.

    Returns:
        A list containing a MatrixTable object for each cell group on the XLSX sheet.
    """
    tables_temp = [sheet_table]
    tables_done: list[MatrixTable] = []

    while len(tables_temp) > 0:
        table = tables_temp.pop().content

        table = _clean_borders(table)

        next_iter = not table
        if next_iter:
            continue

        for i in range(len(table)):
            if _is_empty_line(table, i):
                sub_tab_1 = MatrixTable("", table[:i])
                sub_tab_2 = MatrixTable("", table[i + 1 :])
                tables_temp.extend([sub_tab_2, sub_tab_1])
                next_iter = True
                break

        if next_iter:
            continue

        for j in range(len(table[0])):
            if _is_empty_column(table, j):
                sub_tab_1 = MatrixTable(
                    "",
                    [[table[i][k] for k in range(j)] for i in range(len(table))],
                )
                sub_tab_2 = MatrixTable(
                    "",
                    [
                        [table[i][k] for k in range(j + 1, len(table[0]))]
                        for i in range(len(table))
                    ],
                )
                tables_temp.extend([sub_tab_2, sub_tab_1])
                next_iter = True
                break

        if next_iter:
            continue

        tables_done.append(MatrixTable("", table))

    return tables_done


def extract_subchunks_xlsx(
    path: Path,
    extension: str,
) -> tuple[list[SubChunk], list[MatrixTable], list[str], Path]:
    """Filetype-specific function : extracts subchunks from a XLSX file.

    For XLS & ODS : local file converted to XLSX, subchunks extracted from XLSX.
    Keeps a named sheets structure.
    (ex : "New sheet 'Sheet1' : ... ; End of sheet 'Sheet1'").

    Args:
        path: Path to the local file.
        extension: File extension.

    Returns:
        A tuple containing three lists and a Path:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
        - Path to the processed XLSX file, different from the initial path if a conversion occured.

    """
    filename = path.name
    if extension != "xlsx":
        path = convert_file(path, "xlsx")

    wb = openpyxl.load_workbook(path)

    compiler = ModelCompiler()
    compiler.read_and_parse_archive(str(path))  # type: ignore
    evaluator = Evaluator(compiler.model)

    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []
    image_bank: list[bytes] = []

    for sheet_idx in range(len(wb.worksheets)):
        sheet = wb.worksheets[sheet_idx]
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content=f"** New sheet '{wb.sheetnames[sheet_idx]}' : ",
                page=sheet_idx,
                position=ChunkPosition(0, 0, 0, 0),
                file_name=filename,
            ),
        )

        table = _extract_table_from_sheet(
            wb=wb,
            sheet=sheet,
            sheet_idx=sheet_idx,
            evaluator=evaluator,
        )

        splitted_tables = _split_xlsx_sheet(table)

        for table in splitted_tables:
            compact_matrix(
                table=table,
                do_delete_empty_lines=False,
                do_delete_empty_columns=False,
                do_merge_fake_lines=False,
                do_merge_fake_columns=False,
            )

            is_table = True
            if table.abort():
                subchunk_content = table.read_as_str()
                is_table = False
            else:
                subchunk_content = f"{len(tables)}"
                tables.append(table)
            if subchunk_content.strip():
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=subchunk_content,
                        page=sheet_idx,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                        content_type="table" if is_table else "text",
                    ),
                )

        charts, images_bytes = extract_media_xlsx(path, page_idx=sheet_idx + 1)

        for chart in charts:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=f"{len(tables)}",
                    page=sheet_idx,
                    position=ChunkPosition(0, 0, 0, 0),
                    file_name=filename,
                    content_type="table",
                ),
            )
            tables.append(chart)

        for image_bytes in images_bytes:
            if image_bytes in image_bank:
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=f"{image_bank.index(image_bytes)}",
                        page=sheet_idx,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                        content_type="image",
                    ),
                )
            else:
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=f"{len(images)}",
                        page=sheet_idx,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                        content_type="image",
                    ),
                )
                images.append(f"IMAGE {len(images)} ")
                image_bank.append(image_bytes)

        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content=f" ; End of sheet '{wb.sheetnames[sheet_idx]}' ** ",
                page=sheet_idx,
                position=ChunkPosition(1, 1, 1, 1),
                file_name=filename,
            ),
        )

    wb.close()

    return subchunks, tables, images, path



"""Converts files to Office format or PDF."""

import subprocess
from pathlib import Path
from typing import Literal

from unichunking.utils import logger


def convert_file(
    input_path: Path,
    extension: Literal["docx", "xlsx", "pptx", "pdf"],
) -> Path:
    """Converts DOC/XLS/PPT and ODT/ODS/ODP to DOCX/XLSX/PPTX or PDF.

    Args:
        input_path: Path to the local file to convert.
        extension: Output format.

    Returns:
        The path to the local converted file.
    """
    try:
        subprocess.run(  # noqa: S603
            [
                Path("soffice"),
                "--headless",
                "--convert-to",
                extension,
                input_path,
                "--outdir",
                input_path.parent,
            ],
            check=True,
        )
        logger.debug(
            f"File successfully converted to {extension.upper()} format.",
        )
    except subprocess.CalledProcessError:
        logger.debug(
            "Error during conversion.",
        )

    return input_path.with_suffix(f".{extension}")



"""Functions for handling chunking and pages computing."""

from .chunking import (
    build_chunked_pages,
    split_chunks_with_overlap,
)
from .pages import compute_pages
from .subchunking import extract_subchunks
from .utils import (
    compute_bbox,
    subchunks_to_chunks,
)

__all__ = [
    "build_chunked_pages",
    "compute_bbox",
    "compute_pages",
    "extract_subchunks",
    "split_chunks_with_overlap",
    "subchunks_to_chunks",
]



"""Subchunks extraction for all the supported types."""

from .docx_sub_chunking import extract_subchunks_docx
from .ipynb_sub_chunking import extract_subchunks_ipynb
from .pdf_sub_chunking import extract_subchunks_pdf
from .pptx_sub_chunking import extract_subchunks_pptx
from .txt_sub_chunking import extract_subchunks_txt
from .xlsx_sub_chunking import extract_subchunks_xlsx

__all__ = [
    "extract_subchunks_docx",
    "extract_subchunks_ipynb",
    "extract_subchunks_pdf",
    "extract_subchunks_pptx",
    "extract_subchunks_txt",
    "extract_subchunks_xlsx",
]



"""Extract subchunks from DOCX file."""

import base64
import re
from pathlib import Path
from typing import Any, Callable

from docx import Document
from docx.oxml import parse_xml
from docx.oxml.ns import qn
from docx.table import Table
from docx.text.paragraph import Paragraph

from unichunking.tools import (
    compact_matrix,
    convert_file,
    extract_charts,
)
from unichunking.types import (
    ChunkPosition,
    MatrixTable,
    SubChunk,
)
from unichunking.utils import logger


def _format_toc(toc: list[tuple[str, str]]) -> str:
    text = "Table of content : \n\n\n"

    for number, title in toc:
        try:
            num = int(number)
        except Exception:  # noqa: BLE001, S112
            continue

        text += f"{'#' * num} {title}\n"

    return text


def _extract_toc(para_xml: str) -> tuple[str, str]:
    p = parse_xml(para_xml)
    p_style_val = p.xpath(".//w:pStyle/@w:val")
    if len(p_style_val) > 0:
        title = p_style_val[0]
        numbers = re.findall(r"\d+", title)
        if numbers:
            number = "".join(numbers)
        else:
            return "", ""
    else:
        return "", ""

    w_t_texts = p.xpath(".//w:t/text()")
    w_t_text_combined = ""
    if w_t_texts:
        w_t_text_combined = "".join(w_t_texts)

    return number, w_t_text_combined


def _docx_to_matrixtable(elt: Table) -> MatrixTable:
    matrix_table: list[list[str]] = []
    for row in elt.rows:
        cur_line = [
            cell.text.replace("\n", " ").replace("\xa0", " ") for cell in row.cells
        ]
        matrix_table.append(cur_line)
    return MatrixTable("", matrix_table)


def _compute_position(section: Any) -> ChunkPosition:
    position = ChunkPosition(0, 0, 1, 1)

    if section.page_width:
        if section.left_margin:
            position.x0 = section.left_margin.inches / section.page_width.inches
        if section.right_margin:
            position.x1 = 1 - section.right_margin.inches / section.page_width.inches
    if section.page_height:
        if section.top_margin:
            position.y0 = section.top_margin.inches / section.page_height.inches
        if section.bottom_margin:
            position.y1 = 1 - section.bottom_margin.inches / section.page_height.inches

    return position


def _heading_paragraph(
    para: Paragraph,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    for run in para.runs:
        if "lastRenderedPageBreak" in run.element.xml or (
            "w:br" in run.element.xml and 'type="page"' in run.element.xml
        ):
            cur_page += 1
        text = run.text
        if run.font.bold:
            text = f"**{text}**"
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content=text + " ",
                page=cur_page,
                position=position,
                file_name=filename,
            ),
        )

    return subchunks, cur_page


def _checkbox_paragraph(
    para: Paragraph,
    p: Any,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    seq = """<w:default w:val="1"/>"""

    if seq in p.xml:
        for run in para.runs:
            if "lastRenderedPageBreak" in run.element.xml or (
                "w:br" in run.element.xml and 'type="page"' in run.element.xml
            ):
                cur_page += 1
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content="- [X]" + para.text + " ",
                page=cur_page,
                position=position,
                file_name=filename,
            ),
        )
    else:
        for run in para.runs:
            if "lastRenderedPageBreak" in run.element.xml or (
                "w:br" in run.element.xml and 'type="page"' in run.element.xml
            ):
                cur_page += 1
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content="- [ ]" + para.text + " ",
                page=cur_page,
                position=position,
                file_name=filename,
            ),
        )

    return subchunks, cur_page


def _list_paragraph(
    para: Paragraph,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    for run in para.runs:
        if "lastRenderedPageBreak" in run.element.xml or (
            "w:br" in run.element.xml and 'type="page"' in run.element.xml
        ):
            cur_page += 1
        espaces_indentation = ""
        if para.paragraph_format.left_indent:
            espaces_indentation = " " * 4
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=espaces_indentation + "- " + para.text.strip() + " ",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )
        else:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content="* " + para.text + "\n",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )
    return subchunks, cur_page


def _toc_paragraph(
    para: Paragraph,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
) -> tuple[list[SubChunk], int]:
    if (
        para.style is not None
        and para.style.name is not None
        and "TOC" in para.style.name
    ):
        text = para.text
        if text:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=f"Table des matières : {text}",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )
    return subchunks, cur_page


def _handle_paragraph(
    para: Paragraph,
    para_xml: str,
    position: ChunkPosition,
    subchunks: list[SubChunk],
    cur_page: int,
    filename: str,
    image_rels: dict[str, str],
) -> tuple[list[SubChunk], int]:
    p = parse_xml(para_xml)

    if para.style and para.style.name and para.style.name.lower().startswith("toc"):
        subchunks, cur_page = _toc_paragraph(
            para=para,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    elif para.style and para.style.name and para.style.name.startswith("Heading"):
        subchunks, cur_page = _heading_paragraph(
            para=para,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    elif p.xpath(".//w:checkBox") != []:
        subchunks, cur_page = _checkbox_paragraph(
            para=para,
            p=p,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    elif para.style and para.style.name == "List Paragraph":
        subchunks, cur_page = _list_paragraph(
            para=para,
            position=position,
            subchunks=subchunks,
            cur_page=cur_page,
            filename=filename,
        )

    else:
        for run in para.runs:
            if "lastRenderedPageBreak" in run.element.xml or (
                "w:br" in run.element.xml and 'type="page"' in run.element.xml
            ):
                cur_page += 1
            text = run.text
            if run.font.bold:
                text = f"**{text}**"
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=text.replace("\t", " ") + " ",
                    page=cur_page,
                    position=position,
                    file_name=filename,
                ),
            )

    for drawing in p.xpath(".//w:drawing//a:blip"):
        rid = drawing.get(qn("r:embed"))
        if rid in image_rels:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=image_rels[rid],
                    page=cur_page,
                    position=position,
                    file_name=filename,
                    content_type="image",
                ),
            )

    return subchunks, cur_page


def extract_subchunks_docx(  # noqa: C901, PLR0912
    path: Path,
    extension: str,
    function: Callable[[str], str],
) -> tuple[list[SubChunk], list[MatrixTable], list[str], Path]:
    """Filetype-specific function : extracts subchunks from a DOCX file.

    For DOC & ODT : local file converted to DOCX, subchunks extracted from DOCX.
    Detects any potential page breaks left by renderers to help with further page computing.

    Args:
        path: Path to the local file.
        extension: File extension.
        function: Function to handle images.

    Returns:
        A tuple containing three lists and a Path:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
        - Path to the processed DOCX file, different from the initial path if a conversion occured.
    """
    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []

    filename = path.name

    if extension != "docx":
        path = convert_file(path, "docx")

    doc = Document(str(path))

    toc: list[tuple[str, str]] = []

    img_idx = 0
    image_rels: dict[str, str] = {}
    for rel in doc.part.rels.values():
        if "image" in rel.reltype:
            try:
                image_part = rel.target_part
                image_bytes = image_part.blob
                base64_image = base64.b64encode(image_bytes).decode("utf-8")
                image_description = function(base64_image)
                image_rels[rel.rId] = f"{img_idx}"
                images.append(image_description)
                img_idx += 1
            except Exception as e:  # noqa: BLE001
                logger.debug(f"Image couldn't be processed : {e}")

    cur_page = 0

    for section in doc.sections:
        position = _compute_position(section)

        for elt in section.iter_inner_content():
            if isinstance(elt, Paragraph):
                para_xml = elt._p.xml  # type: ignore # noqa: SLF001
                if 'w:name="_Toc' in para_xml:
                    number, title = _extract_toc(
                        para_xml=para_xml,
                    )
                    if number and title:
                        toc.append((number, title))

                if "<c:chart " in para_xml:
                    for plot in extract_charts(
                        file_path=path,
                        element_xml=para_xml,
                        extension="docx",
                    ):
                        subchunks.append(
                            SubChunk(
                                subchunk_id=len(subchunks),
                                content=f"{len(tables)}",
                                page=cur_page,
                                position=position,
                                file_name=filename,
                                content_type="table",
                            ),
                        )
                        tables.append(plot)

                subchunks, cur_page = _handle_paragraph(
                    para=elt,
                    para_xml=para_xml,
                    position=position,
                    subchunks=subchunks,
                    cur_page=cur_page,
                    filename=filename,
                    image_rels=image_rels,
                )

            else:
                table = compact_matrix(table=_docx_to_matrixtable(elt))
                if table.abort():
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=table.read_as_str(),
                            page=cur_page,
                            position=position,
                            file_name=filename,
                        ),
                    )
                else:
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=f"{len(tables)}",
                            page=cur_page,
                            position=position,
                            file_name=filename,
                            content_type="table",
                        ),
                    )
                    tables.append(table)

    toc_text = _format_toc(toc)

    subchunks.append(
        SubChunk(
            subchunk_id=len(subchunks),
            content=toc_text,
            page=1,
            position=ChunkPosition(0, 0, 1, 1),
            file_name=filename,
            content_type="text",
        ),
    )

    subchunks = [s for s in subchunks if s.content.strip()]

    return subchunks, tables, images, path



"""Global multi-format chunk extraction function."""

from .formating import (
    build_chunked_pages,
    compute_bbox,
    compute_pages,
    extract_subchunks,
    split_chunks_with_overlap,
    subchunks_to_chunks,
)
from .settings import unisettings
from .tools import convert_file
from .types import StatusManager

__all__ = [
    "StatusManager",
    "build_chunked_pages",
    "compute_bbox",
    "compute_pages",
    "convert_file",
    "extract_subchunks",
    "split_chunks_with_overlap",
    "subchunks_to_chunks",
    "unisettings",
]



"""Useful functions for handling SubChunks and Chunks."""

from uuid import uuid4

from llmax import tokens

from unichunking.types import Chunk, SubChunk


def subchunks_to_chunks(
    subchunks: list[SubChunk],
    file_hash: str = "",
) -> list[Chunk]:
    """Convert a SubChunk object to a Chunk object with the same content."""
    return [
        Chunk(
            chunk_num=i + 1,
            content=subchunks[i].content,
            metadata={
                "page": subchunks[i].page,
                "file_name": subchunks[i].file_name,
                "positions": [subchunks[i].position.to_dict()],
            },
            embedding=[0] * 1536,
            chunk_uid=uuid4(),
            file_hash=file_hash,
            num_tokens=tokens.count(subchunks[i].content),
        )
        for i in range(len(subchunks))
    ]


def compute_bbox(
    chunks: list[Chunk],
) -> dict[str, float]:
    """Compute the coordinates of the minimal bounding box for a list of chunks."""
    min_x0, min_y0, max_x1, max_y1 = 1, 1, 0, 0

    for chunk in chunks:
        positions = [chunk.metadata.get("bbox", [])]
        if not positions[0]:
            positions = chunk.metadata["positions"]
        for chunk_position in positions:
            min_x0 = max(0, min(min_x0, chunk_position["x0"]))
            min_y0 = max(0, min(min_y0, chunk_position["y0"]))
            max_x1 = min(1, max(max_x1, chunk_position["x1"]))
            max_y1 = min(1, max(max_y1, chunk_position["y1"]))

    return {
        "x0": min_x0,
        "y0": min_y0,
        "x1": max_x1,
        "y1": max_y1,
    }



"""Final return type for chunk objects."""

from typing import Any
from uuid import UUID


class Chunk:
    """Class for storing a piece of text and associated metadata.

    Fields:
        - chunk_num: The number of chunk inside the page.
        - content: Text of chunk.
        - metadata: {
                page : Number of page to which it belongs,
                file_name: Name of file,
                positions: Positions of all SubChunks inside the Chunk,
                bbox: Coordinates of the minimal bounding box for the Chunk, only exists if computed with dedicated function,
            },
        - embedding: Embedding of chunk text (not handled by unichunking package).
        - chunk_uid: Unique identifier of the chunk.
        - file_hash: File unique reference.
        - num_tokens: Number of tokens.
    """

    def __init__(
        self: "Chunk",
        chunk_num: int,
        content: str,
        metadata: dict[str, Any],
        embedding: list[float],
        chunk_uid: UUID,
        file_hash: str,
        num_tokens: int,
    ) -> None:
        """Create a Chunk object."""
        self.chunk_num = chunk_num
        self.content = content
        self.metadata = metadata
        self.embedding = embedding
        self.chunk_uid = chunk_uid
        self.file_hash = file_hash
        self.num_tokens = num_tokens

    def to_dict(
        self: "Chunk",
        as_subchunk: bool = False,
    ) -> dict[str, str | int | dict[str, dict[str, float] | list[dict[str, float]]]]:
        """Converts a Chunk object to a dictionary."""
        output_dict = {
            "content": self.content,
            "size": self.num_tokens,
            "page": self.metadata["page"],
        }

        if as_subchunk:
            output_dict["position"] = self.metadata["positions"][0]

        else:
            output_dict["position"] = {
                "bbox": self.metadata["bbox"],
                "subchunks": self.metadata["positions"],
            }
        return output_dict



"""Extract subchunks from PPTX file."""

import base64
import contextlib
from pathlib import Path
from typing import Any, Callable

from pptx import Presentation

from unichunking.tools import (
    compact_matrix,
    convert_file,
    extract_charts,
)
from unichunking.types import (
    ChunkPosition,
    MatrixTable,
    SubChunk,
)
from unichunking.utils import logger

MSO_SHAPE_TYPE_GROUP = 6


def _intersect(
    shape_1_pos: ChunkPosition,
    shape_2_pos: ChunkPosition,
) -> bool:
    shape_1_bary = (
        (shape_1_pos.x0 + shape_1_pos.x1) / 2,
        (shape_1_pos.y0 + shape_1_pos.y1) / 2,
    )
    shape_2_bary = (
        (shape_2_pos.x0 + shape_2_pos.x1) / 2,
        (shape_2_pos.y0 + shape_2_pos.y1) / 2,
    )
    return (
        (shape_1_pos.x0 <= shape_2_bary[0] <= shape_1_pos.x1)
        and (shape_1_pos.y0 <= shape_2_bary[1] <= shape_1_pos.y1)
    ) or (
        (shape_2_pos.x0 <= shape_1_bary[0] <= shape_2_pos.x1)
        and (shape_2_pos.y0 <= shape_1_bary[1] <= shape_2_pos.y1)
    )


def _create_shape_groups(
    shapes: list[Any],
    dimensions: ChunkPosition,
) -> list[tuple[ChunkPosition, list[tuple[Any, ChunkPosition]]]]:
    formatted_shapes_info: list[
        tuple[ChunkPosition, list[tuple[Any, ChunkPosition]]]
    ] = []
    while shapes:
        shape = shapes.pop()
        position = ChunkPosition(
            (shape.left - dimensions.x0) / (dimensions.x1 - dimensions.x0),
            (shape.top - dimensions.y0) / (dimensions.y1 - dimensions.y0),
            (shape.left + shape.width - dimensions.x0)
            / (dimensions.x1 - dimensions.x0),
            (shape.top + shape.height - dimensions.y0)
            / (dimensions.y1 - dimensions.y0),
        )
        formatted_shapes_info.append((position, [(shape, position)]))

    changed_something = True
    while changed_something:
        changed_something = False
        for i in range(len(formatted_shapes_info)):
            for j in range(len(formatted_shapes_info)):
                if i == j:
                    continue
                if _intersect(formatted_shapes_info[i][0], formatted_shapes_info[j][0]):
                    position = ChunkPosition(
                        x0=min(
                            formatted_shapes_info[i][0].x0,
                            formatted_shapes_info[j][0].x0,
                        ),
                        y0=min(
                            formatted_shapes_info[i][0].y0,
                            formatted_shapes_info[j][0].y0,
                        ),
                        x1=max(
                            formatted_shapes_info[i][0].x1,
                            formatted_shapes_info[j][0].x1,
                        ),
                        y1=max(
                            formatted_shapes_info[i][0].y1,
                            formatted_shapes_info[j][0].y1,
                        ),
                    )
                    shapes_list = formatted_shapes_info[i][1]
                    shapes_list.extend(formatted_shapes_info[j][1])
                    formatted_shapes_info[i] = (position, shapes_list)
                    formatted_shapes_info.pop(j)
                    changed_something = True
                    break
            if changed_something:
                break

    return formatted_shapes_info


def _insert_in_order(
    position: ChunkPosition,
    shape_list: list[Any],
    ordered_shape_groups: list[tuple[ChunkPosition, list[Any]]],
    verticality_bias: float = 0.125,
) -> list[tuple[ChunkPosition, list[Any]]]:
    coord = ((position.x0 + position.x1) * (1 - verticality_bias) / 2) ** 2 + (
        (position.y0 + position.y1) * (1 + verticality_bias) / 2
    ) ** 2
    i = 0
    while i < len(ordered_shape_groups):
        if (
            coord
            < (
                (ordered_shape_groups[i][0].x0 + ordered_shape_groups[i][0].x1)
                * (1 - verticality_bias)
                / 2
            )
            ** 2
            + (
                (ordered_shape_groups[i][0].y0 + ordered_shape_groups[i][0].y1)
                * (1 + verticality_bias)
                / 2
            )
            ** 2
        ):
            return (
                ordered_shape_groups[:i]
                + [(position, shape_list)]
                + ordered_shape_groups[i:]
            )
        i += 1
    return [*ordered_shape_groups, (position, shape_list)]


def _sort_shapes(
    shapes: list[Any],
    dimensions: ChunkPosition,
) -> list[Any]:
    """Sorts the shapes on a slide to emulate a natural reading order.

    Overlapping shapes are grouped, and the groups are sorted by increasing distance from their barycenter to the top left corner of the slide.
    Then the shapes in each group are sorted recursively, the coordinates of the group replacing those of the slide, etc.
    The parameter vertical_bias in insert_in_order gives a stronger weight to the vertical axis than to the horizontal one when computing distances.

    Args:
        shapes: List of shapes to sort.
        dimensions: The dimensions of the plane on which to sort the shapes (the slide, or a shape group).

    Returns:
        A list of ordered shapes.
    """
    if len(shapes) <= 1:
        if shapes and shapes[0].shape_type == MSO_SHAPE_TYPE_GROUP:
            return _sort_shapes(list(shapes[0].shapes), dimensions)
        return shapes

    shape_groups = _create_shape_groups(shapes, dimensions)

    ordered_shape_groups: list[tuple[ChunkPosition, list[Any]]] = []

    if len(shape_groups) == 1:
        shape_group = shape_groups[0]
        shapes_in_group = shape_group[1]
        for shape_info in shapes_in_group:
            shape = shape_info[0]
            shape_position = shape_info[1]
            if shape.shape_type == MSO_SHAPE_TYPE_GROUP:
                ordered_shape_groups = _insert_in_order(
                    position=shape_position,
                    shape_list=_sort_shapes(list(shape.shapes), shape_position),
                    ordered_shape_groups=ordered_shape_groups,
                )
            else:
                ordered_shape_groups = _insert_in_order(
                    position=shape_position,
                    shape_list=[shape],
                    ordered_shape_groups=ordered_shape_groups,
                )

    else:
        for shape_group in shape_groups:
            group_position = shape_group[0]
            shapes_in_group_info = shape_group[1]
            shapes_in_group = [
                shape_in_group_info[0] for shape_in_group_info in shapes_in_group_info
            ]
            ordered_shape_groups = _insert_in_order(
                position=group_position,
                shape_list=_sort_shapes(shapes_in_group, group_position),
                ordered_shape_groups=ordered_shape_groups,
            )

    ordered_shapes: list[Any] = []
    for shape_group in ordered_shape_groups:
        ordered_shapes.extend(shape_group[1])

    return ordered_shapes


def _handle_text_frame(
    shape: Any,
    subchunks: list[SubChunk],
    slide_idx: int,
    position: ChunkPosition,
    filename: str,
) -> list[SubChunk]:
    text_frame = shape.text_frame
    for paragraph in text_frame.paragraphs:
        for run in paragraph.runs:
            if run.text:
                text = run.text
                if run.font and run.font.bold:
                    text = f"**{text}**"
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=text + " ",
                        page=slide_idx,
                        position=position,
                        file_name=filename,
                    ),
                )
    return subchunks


def _handle_charts_and_tables(
    shape: Any,
    subchunks: list[SubChunk],
    tables: list[MatrixTable],
    slide_idx: int,
    position: ChunkPosition,
    filename: str,
    file_path: Path,
) -> tuple[list[SubChunk], list[MatrixTable]]:
    if shape.has_chart:
        for plot in extract_charts(
            file_path,
            shape._element.xml,  # noqa: SLF001
            "pptx",
            slide_idx + 1,
        ):
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=f"{len(tables)}",
                    page=slide_idx,
                    position=position,
                    file_name=filename,
                    content_type="table",
                ),
            )
            tables.append(plot)

    if shape.has_table:
        table = shape.table

        matrix_table = [
            [str(table.cell(i, j).text) for j in range(len(table.columns))]
            for i in range(len(table.rows))
        ]
        matrix_table = compact_matrix(table=MatrixTable("", matrix_table))
        if matrix_table.abort():
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=matrix_table.read_as_str(),
                    page=slide_idx,
                    position=position,
                    file_name=filename,
                ),
            )
        else:
            subchunks.append(
                SubChunk(
                    subchunk_id=len(subchunks),
                    content=f"{len(tables)}",
                    page=slide_idx,
                    position=position,
                    file_name=filename,
                    content_type="table",
                ),
            )
            tables.append(matrix_table)

    return subchunks, tables


def extract_subchunks_pptx(
    path: Path,
    extension: str,
    function: Callable[[str], str],
) -> tuple[list[SubChunk], list[MatrixTable], list[str], Path]:
    """Filetype-specific function : extracts subchunks from a PPTX file.

    For PPT & ODP : local file converted to PPTX, subchunks extracted from PPTX.
    Keeps a numbered slides structure.
    (ex : "Slide 1 : ... ; End of slide 1").

    Args:
        path: Path to the local file.
        extension: File extension.
        function: Function to handle images.

    Returns:
        A tuple containing three lists and a Path:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
        - Path to the processed PPTX file, different from the initial path if a conversion occured.
    """
    if extension != "pptx":
        path = convert_file(path, "pptx")

    prs = Presentation(str(path))

    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []

    for slide_idx in range(len(prs.slides)):
        slide = prs.slides[slide_idx]
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content=f"** Slide {slide_idx + 1} : ",
                page=slide_idx,
                position=ChunkPosition(0, 0, 0, 0),
                file_name=path.name,
            ),
        )

        slide_width = float(prs.slide_width or 1)
        slide_height = float(prs.slide_height or 1)
        ordered_shapes = _sort_shapes(
            shapes=list(slide.shapes),
            dimensions=ChunkPosition(0, 0, slide_width, slide_height),
        )

        for shape in ordered_shapes:
            position = ChunkPosition(
                shape.left / slide_width,
                shape.top / slide_height,
                (shape.left + shape.width) / slide_width,
                (shape.top + shape.height) / slide_height,
            )

            subchunks, tables = _handle_charts_and_tables(
                shape=shape,
                subchunks=subchunks,
                tables=tables,
                slide_idx=slide_idx,
                position=position,
                filename=path.name,
                file_path=path,
            )

            if shape.has_text_frame:
                subchunks = _handle_text_frame(
                    shape=shape,
                    subchunks=subchunks,
                    slide_idx=slide_idx,
                    position=position,
                    filename=path.name,
                )

            with contextlib.suppress(Exception):
                image = shape.image
                image_bytes = image.blob

                base64_image = base64.b64encode(image_bytes).decode("utf-8")
                try:
                    image_description = function(base64_image)
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=f"{len(images)}",
                            page=slide_idx,
                            position=position,
                            file_name=path.name,
                            content_type="image",
                        ),
                    )
                    images.append(image_description)
                except Exception as e:  # noqa: BLE001
                    logger.debug(f"Image couldn't be processed : {e}")

        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content=f" ; End of slide {slide_idx + 1} ** ",
                page=slide_idx,
                position=ChunkPosition(
                    1,
                    1,
                    1,
                    1,
                ),
                file_name=path.name,
            ),
        )

    return subchunks, tables, images, path
